import copy
import os
import pathlib

import yaml
from aiida.engine import ExitCode, ToContext, WorkChain
from aiida.orm import Bool, Code, Dict, Int, List, SinglefileData, Str, StructureData
from aiida.plugins import WorkflowFactory

from aiida_nanotech_empa.utils import common_utils
from aiida_nanotech_empa.workflows.cp2k.cp2k_utils import (
    determine_kinds,
    dict_merge,
    get_cutoff,
    get_kinds_section,
)

Cp2kBaseWorkChain = WorkflowFactory("cp2k.base")


class Cp2kBulkOptWorkChain(WorkChain):
    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input("code", valid_type=Code)
        spec.input("structure", valid_type=StructureData)

        spec.input(
            "charge",  # +1 means one electron removed
            valid_type=Int,
            default=lambda: Int(0),
            required=False,
        )
        spec.input(
            "fixed_atoms", valid_type=Str, default=lambda: Str(""), required=False
        )
        spec.input(
            "cell_opt", valid_type=Bool, default=lambda: Bool(False), required=False
        )
        spec.input(
            "symmetry",
            valid_type=Str,
            default=lambda: Str("ORTHORHOMBIC"),
            required=False,
        )
        spec.input(
            "cell_freedom",
            valid_type=Str,
            default=lambda: Str("KEEP_SYMMETRY"),
            required=False,
        )
        spec.input(
            "multiplicity", valid_type=Int, default=lambda: Int(0), required=False
        )
        spec.input(
            "magnetization_per_site",
            valid_type=List,
            default=lambda: List(list=[]),
            required=False,
        )
        spec.input("vdw", valid_type=Bool, default=lambda: Bool(False), required=False)
        spec.input(
            "protocol",
            valid_type=Str,
            default=lambda: Str("standard"),
            required=False,
            help="Settings to run simulations with.",
        )
        spec.input(
            "options",
            valid_type=dict,
            non_db=True,
            required=False,
            help="Define options for the cacluations: walltime, memory, CPUs, etc.",
        )

        # workchain outline
        spec.outline(cls.setup, cls.submit_calc, cls.finalize)
        spec.outputs.dynamic = True

        spec.exit_code(
            390,
            "ERROR_TERMINATION",
            message="One or more steps of the work chain failed.",
        )

    def setup(self):
        self.report("Inspecting input and setting up things")

        # --------------------------------------------------

    def submit_calc(self):

        # load input template
        the_protocol = "./protocols/bulk_opt_protocol.yml"
        if self.inputs.cell_opt.value:
            the_protocol = "./protocols/cell_opt_protocol.yml"
        with open(
            pathlib.Path(__file__).parent / the_protocol, encoding="utf-8"
        ) as handle:
            protocols = yaml.safe_load(handle)
            input_dict = copy.deepcopy(protocols[self.inputs.protocol.value])

        structure = self.inputs.structure
        # cutoff
        self.ctx.cutoff = get_cutoff(structure=structure)

        # get initial magnetization
        magnetization_per_site = copy.deepcopy(self.inputs.magnetization_per_site)
        structure_with_tags, kinds_dict = determine_kinds(
            structure, magnetization_per_site
        )

        ase_atoms = structure_with_tags.get_ase()

        builder = Cp2kBaseWorkChain.get_builder()
        builder.cp2k.code = self.inputs.code
        builder.cp2k.structure = StructureData(ase=ase_atoms)
        builder.cp2k.file = {
            "basis": SinglefileData(
                file=os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    ".",
                    "data",
                    "BASIS_MOLOPT",
                )
            ),
            "pseudo": SinglefileData(
                file=os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    ".",
                    "data",
                    "POTENTIAL",
                )
            ),
        }

        # charge
        input_dict["FORCE_EVAL"]["DFT"]["CHARGE"] = self.inputs.charge.value
        # vdw
        if not self.inputs.vdw.value:
            input_dict["FORCE_EVAL"]["DFT"]["XC"].pop("VDW_POTENTIAL")

        # UKS
        if self.inputs.multiplicity.value > 0:
            input_dict["FORCE_EVAL"]["DFT"]["UKS"] = ".TRUE."
            input_dict["FORCE_EVAL"]["DFT"][
                "MULTIPLICITY"
            ] = self.inputs.multiplicity.value

        # only bulk opt
        if not self.inputs.cell_opt.value:
            # fixed atoms
            input_dict["MOTION"]["CONSTRAINT"]["FIXED_ATOMS"][
                "LIST"
            ] = self.inputs.fixed_atoms.value

        # cell opt
        else:
            # cell symmetry
            input_dict["FORCE_EVAL"]["SUBSYS"]["CELL"][
                "SYMMETRY"
            ] = self.inputs.symmetry.value

            # cell do free
            if self.inputs.cell_freedom.value == "KEEP_SYMMETRY":
                input_dict["MOTION"]["CELL_OPT"]["KEEP_SYMMETRY"] = ""
            elif self.inputs.cell_freedom.value == "KEEP_ANGLES":
                input_dict["MOTION"]["CELL_OPT"]["KEEP_ANGLES"] = ""

        # cutoff
        input_dict["FORCE_EVAL"]["DFT"]["MGRID"]["CUTOFF"] = self.ctx.cutoff

        # KINDS section
        self.ctx.kinds_section = get_kinds_section(kinds_dict, protocol="gpw")
        dict_merge(input_dict, self.ctx.kinds_section)

        # Setup options.
        if "options" in self.inputs:
            builder.cp2k.metadata.options = self.inputs.options

        # Setup walltime.
        if "max_wallclock_seconds" in self.inputs.options:
            input_dict["GLOBAL"]["WALLTIME"] = max(
                self.inputs.options["max_wallclock_seconds"] - 600, 600
            )

        # parser
        builder.cp2k.metadata.options.parser_name = "cp2k_advanced_parser"

        # handlers

        builder.handler_overrides = Dict(
            {"restart_incomplete_calculation": {"enabled": True}}
        )

        # cp2k input dictionary
        builder.cp2k.parameters = Dict(input_dict)

        submitted_node = self.submit(builder)
        return ToContext(opt=submitted_node)

    def finalize(self):
        self.report("Finalizing...")

        if not self.ctx.opt.is_finished_ok:
            return self.exit_codes.ERROR_TERMINATION

        for out in self.ctx.opt.outputs:
            self.out(out, self.ctx.opt.outputs[out])

        # Add extras
        struc = self.ctx.opt.outputs.output_structure
        ase_geom = struc.get_ase()
        self.node.base.extras.set(
            "thumbnail", common_utils.thumbnail(ase_struc=ase_geom)
        )
        self.node.base.extras.set("formula", struc.get_formula())

        return ExitCode(0)
